from typing import List, Tuple, Dict
import gymnasium as gym
import numpy as np
from scipy.stats import poisson


class JacksCarRental(gym.Env):
    """
    Jack's Car Rental environment
    """
    def __init__(
        self,
        max_cars: int = 20,
        max_move_cars: int = 5,
        rental_credit: int = 10,
        move_car_cost: int = 2,
        request_lambda: List[int] = [3, 4],
        return_lambda: List[int] = [3, 2],
        init_state_option: str = "random",
        poisson_max_value: int = 12,
    ) -> None:
        """
        Initializes the environment with the specified parameters.

        Args:
            max_cars: The maximum number of cars in the environment. Defaults to 20.
            max_move_cars: The maximum number of cars that can be moved in a single action. Defaults to 5.
            rental_credit: The credit earned for renting a car. Defaults to 10.
            move_car_cost: The cost of moving a car. Defaults to 2.
            request_lambda: The lambda values for the Poisson distribution of car rental requests. Defaults to [3, 4].
            return_lambda: The lambda values for the Poisson distribution of car returns. Defaults to [3, 2].
            init_state_option: The option for initializing the initial state. Defaults to "random".
            poisson_max_value: The maximum value for the Poisson distribution. Defaults to 12.

        Returns:
            None
        """

        self.max_cars = max_cars
        self.max_move_cars = max_move_cars
        self.rental_credit = rental_credit
        self.move_car_cost = move_car_cost
        self.request_lambda = request_lambda
        self.return_lambda = return_lambda

        self.action_space = gym.spaces.Discrete(2 * max_move_cars + 1)
        self.observation_space = gym.spaces.Tuple(
            (gym.spaces.Discrete(max_cars + 1), gym.spaces.Discrete(max_cars + 1))
        )

        self._precompute_poisson_probabilities(max_value=poisson_max_value)
        self.reset(init_state_option)

    def _move_cars_and_calculate_cost(
        self, state: Tuple[int, int], action: int
    ) -> Tuple[Tuple[int, int], int]:
        """
        Move cars between locations and calculate the cost.

        Args:
            state: A tuple containing the number of cars at location A and location B.
            action: The number of cars to move from A to B or from B to A.

        Returns:
            A tuple containing the new state after moving the cars, with the number of cars at location A and location B.
            The cost of moving the cars.
        """
        cars_at_A, cars_at_B = state
        cars_moved = min(
            cars_at_A if action > 0 else cars_at_B, abs(action), self.max_move_cars
        )
        cars_at_A -= cars_moved * np.sign(action)
        cars_at_B += cars_moved * np.sign(action)
        return (max(cars_at_A, 0), max(cars_at_B, 0)), abs(action) * self.move_car_cost

    def _simulate_rentals_and_returns(
        self, state: Tuple[int, int]
    ) -> Tuple[Tuple[int, int], int]:
        """
        Simulates the process of rentals and returns of cars at two locations.

        Parameters:
            state (tuple): A tuple containing the number of cars at location A and location B.

        Returns:
            tuple: A tuple containing the updated number of cars at location A and location B after the rentals and returns, and the reward earned from the rentals.
        """
        cars_at_A, cars_at_B = state
        requests_A = poisson.rvs(self.request_lambda[0])
        requests_B = poisson.rvs(self.request_lambda[1])
        returns_A = poisson.rvs(self.return_lambda[0])
        returns_B = poisson.rvs(self.return_lambda[1])

        rentals_A = min(cars_at_A, requests_A)
        rentals_B = min(cars_at_B, requests_B)

        reward = (rentals_A + rentals_B) * self.rental_credit

        cars_at_A = min(cars_at_A - rentals_A + returns_A, self.max_cars)
        cars_at_B = min(cars_at_B - rentals_B + returns_B, self.max_cars)

        return (cars_at_A, cars_at_B), reward

    def step(self, action: int) -> Tuple[Dict[int, int], float, bool, bool, dict]:
        """
        Performs a single step in the environment.

        Args:
            action: The action to take.

        Returns:
            A tuple containing the new state, the reward, and two boolean values indicating whether the episode is done and whether the state is terminal. The new state is represented as a dictionary. The reward is a float value. The two boolean values are False by default.
        """
        action = action - self.max_move_cars
        moved_state, move_cost = self._move_cars_and_calculate_cost(self.state, action)
        new_state, rental_reward = self._simulate_rentals_and_returns(moved_state)
        reward = rental_reward - move_cost
        self.state = new_state
        return new_state, reward, False, False, {}

    def reset(self, option: str = "random"):
        """
        Resets the state of the environment.

        Parameters:
            option (str, optional): The reset option. Can be "random" or "equal". Defaults to "random".

        Returns:
            tuple: The new state of the environment after the reset and a dictionary containing the reset option.
        """

        if option == "random":
            self.state = (
                np.random.randint(0, self.max_cars + 1),
                np.random.randint(0, self.max_cars + 1),
            )
        elif option == "equal":
            initial_cars = self.max_cars // 2
            self.state = (initial_cars, initial_cars)
        return self.state, {"Reset Option": option}

    def render(self, mode: str = "human") -> str:
        """
        Render the current state of the environment.

        Args:
            mode (str): The mode in which to render the environment.
                Must be one of "human" or "ansi". Defaults to "human".

        Returns:
            str: The rendered state of the environment in the specified mode.

        Raises:
            NotImplementedError: If the specified mode is not supported.
        """
        if mode == "human":
            print(f"Location A: {self.state[0]} cars, Location B: {self.state[1]} cars")
        elif mode == "ansi":
            return f"Location A: {self.state[0]} cars, Location B: {self.state[1]} cars"
        else:
            raise NotImplementedError("Render mode not supported: " + mode)

    def _precompute_poisson_probabilities(self, max_value: int = 12) -> None:
        """
        Precomputes the Poisson probabilities for the given maximum value.

        Args:
            max_value: The maximum value for the Poisson distribution. Defaults to 12.

        Returns:
            None
        """

        self.poisson_probs: Dict[str, Dict[str, List[float]]] = {
            "request": {
                "A": [
                    poisson.pmf(i, self.request_lambda[0]) for i in range(max_value + 1)
                ],
                "B": [
                    poisson.pmf(i, self.request_lambda[1]) for i in range(max_value + 1)
                ],
            },
            "return": {
                "A": [
                    poisson.pmf(i, self.return_lambda[0]) for i in range(max_value + 1)
                ],
                "B": [
                    poisson.pmf(i, self.return_lambda[1]) for i in range(max_value + 1)
                ],
            },
        }

    def _get_probability_of_rental_requests(
        self, request_A: int, request_B: int
    ) -> float:
        """
        Calculates the probability of rental requests occurring simultaneously.

        Args:
            request_A (int): The number of rental requests from source A.
            request_B (int): The number of rental requests from source B.

        Returns:
            float: The probability of both rental requests occurring.
        """
        prob_A = self.poisson_probs["request"]["A"][request_A]
        prob_B = self.poisson_probs["request"]["B"][request_B]

        return prob_A * prob_B

    def _get_probability_of_car_returns(self, return_A: int, return_B: int) -> float:
        """Calculate the probability of both car returns happening simultaneously.

        Args:
            return_A: The number of car returns at location A.
            return_B: The number of car returns at location B.

        Returns:
            The probability of both car returns happening simultaneously.
        """
        return (
            self.poisson_probs["return"]["A"][return_A]
            * self.poisson_probs["return"]["B"][return_B]
        )

    def build_mdp(self):
        """
        Generates the transition and reward dictionary for the Markov Decision Process (MDP).

        This method iterates over all possible states, actions, and events to calculate the
        transition probabilities and rewards for the MDP. It takes into account the number
        of cars at locations A and B, the number of cars moved between locations, the number
        of cars rented and returned at each location, and the associated probabilities of
        these events.

        Returns:
            t_r_dict (dict): A dictionary containing the transition probabilities and rewards
                for each state and action pair. The keys are tuples representing the state and
                action, and the values are tuples representing the next state, the probability
                of transitioning to the next state, and the reward associated with the
                transition.
        """
        t_r_dict = {}

        for cars_at_A in range(self.max_cars + 1):
            for cars_at_B in range(self.max_cars + 1):
                state = (cars_at_A, cars_at_B)

                for action in range(-self.max_move_cars, self.max_move_cars + 1):
                    moved_state, move_cost = self._move_cars_and_calculate_cost(
                        state, action
                    )

                    for request_A in range(0, self.max_cars + 1):
                        for request_B in range(0, self.max_cars + 1):
                            prob_request = (
                                self.poisson_probs["request"]["A"][request_A]
                                * self.poisson_probs["request"]["B"][request_B]
                            )
                            rentals_A = min(moved_state[0], request_A)
                            rentals_B = min(moved_state[1], request_B)
                            rental_reward = (rentals_A + rentals_B) * self.rental_credit

                            for return_A in range(0, self.max_cars + 1):
                                for return_B in range(0, self.max_cars + 1):
                                    prob_return = (
                                        self.poisson_probs["return"]["A"][return_A]
                                        * self.poisson_probs["return"]["B"][return_B]
                                    )
                                    next_state = (
                                        min(
                                            moved_state[0] - rentals_A + return_A,
                                            self.max_cars,
                                        ),
                                        min(
                                            moved_state[1] - rentals_B + return_B,
                                            self.max_cars,
                                        ),
                                    )
                                    prob_total = prob_request * prob_return
                                    reward = rental_reward - move_cost

                                    if (state, action) in t_r_dict:
                                        (
                                            existing_state,
                                            existing_reward,
                                            existing_done,
                                        ) = t_r_dict[(state, action)]
                                        t_r_dict[(state, action)] = (
                                            existing_state,
                                            existing_reward + prob_total * reward,
                                            existing_done,
                                        )
                                    else:
                                        t_r_dict[(state, action)] = (
                                            next_state,
                                            prob_total * reward,
                                            False,
                                        )

        return t_r_dict



import numpy as np
from ..constants import *


def generate_ellipse(
    rows: int,
    cols: int,
    a: int,
    b: int,
    center_x: int,
    center_y: int,
    value_true: float = PATH,
    value_false: float = WALL,
) -> np.ndarray:
    """
    Generate an ellipse shape in a 2D grid.

    Parameters:
        rows (int): The number of rows in the grid.
        cols (int): The number of columns in the grid.
        a (int): The horizontal radius of the ellipse.
        b (int): The vertical radius of the ellipse.
        center_x (int): The x-axis offset of the ellipse.
        center_y (int): The y-axis offset of the ellipse.
        value_true (float): The value to assign to the center of the ellipse.
        value_false (float): The value to assign to the outside of the ellipse.

    Returns:
        np.ndarray: An array representing the generated ellipse shape in the grid.
    """
    ellipse = np.ones((rows, cols), dtype=int) * value_false

    for i in range(rows):
        for j in range(cols):
            y, x = i - center_y, j - center_x

            if (x**2 / a**2) + (y**2 / b**2) <= 1:
                ellipse[i][j] = value_true

    return ellipse


def clamp(value, min_value: float, max_value: float) -> float:
    """
    Clamp a value between a minimum and maximum value.

    Parameters:
        value (float): The value to be clamped.
        min_value (float): The minimum value that the value parameter can be.
        max_value (float): The maximum value that the value parameter can be.

    Returns:
        float: The clamped value.

    """
    return max(min_value, min(value, max_value))



from typing import Tuple
import copy

from ..constants import Action
from ..display.player import PlayerDisplayer


class Player:
    def __init__(self, start_position: Tuple[int, int] = None):
        self._position = list(start_position) if start_position else None
        self._rendered_position = list(start_position) if start_position else None

        self.movement_speed = 0.25  # Adjust for desired speed
        self.moving = False
        self.face_orientation = Action.LEFT  # default sprite looks left

        self._displayer = None

    def set_position(self, position):
        self.position = position
        self.rendered_position = position
        
    def __deepcopy__(self, memo):
        new_player = copy.copy(self)
        memo[id(self)] = new_player
        new_player._displayer = None
        return new_player

    @property
    def displayer(self):
        if self._displayer is None:
            self._displayer = PlayerDisplayer(self)
        return self._displayer

    @property
    def position(self):
        if self._position is None:
            return None
        return tuple(self._position)

    @position.setter
    def position(self, value):
        if isinstance(value, tuple):
            self._position = list(value)
        else:
            self._position = value

    @property
    def rendered_position(self):
        if self._rendered_position is None:
            return None
        return tuple(self._rendered_position)

    @rendered_position.setter
    def rendered_position(self, value):
        if isinstance(value, tuple):
            self._rendered_position = list(value)
        else:
            self._rendered_position = value

    def potential_next_position(self, action: Action) -> Tuple[int, int]:
        """
        Calculate the potential next position based on the current position and the specified action.

        Parameters:
            action (Action): The action to take.

        Returns:
            Tuple[int, int]: The potential next position.
        """
        potential_position = list(self.position)

        if action == Action.UP:  # Up
            potential_position[0] -= 1
        elif action == Action.RIGHT:  # Right
            potential_position[1] += 1
        elif action == Action.DOWN:  # Down
            potential_position[0] += 1
        elif action == Action.LEFT:  # Left
            potential_position[1] -= 1

        return tuple(potential_position)

    def move_render_position(self) -> None:
        """
        Updates the rendered position of the object based on its current position and movement speed.

        Parameters:
            None.

        Returns:
            None.
        """
        new_rendered_position = list(self._rendered_position)

        for i in range(2):  # For x and y coordinates
            diff = self._position[i] - new_rendered_position[i]
            if abs(diff) > 0.01:  # Adjust for desired precision
                new_rendered_position[i] += diff * self.movement_speed

        self._rendered_position = new_rendered_position

    def _positions_are_close(
        self, pos1: Tuple[int, int], pos2: Tuple[int, int], threshold: float = 0.025
    ) -> bool:
        """
        Check if the given positions are close to each other within a threshold.

        Args:
            pos1 (Tuple[int, int]): The first position.
            pos2 (Tuple[int, int]): The second position.
            threshold (float, optional): The threshold value for closeness. Defaults to 0.025.

        Returns:
            bool: True if the positions are close, False otherwise.
        """
        return abs(pos1[0] - pos2[0]) < threshold and abs(pos1[1] - pos2[1]) < threshold



import numpy as np
import random
from collections import deque
from queue import PriorityQueue
from typing import List, Tuple, Union, Optional, Dict

from ..constants import WALL, PATH, START, TARGET, CorridorMoveStatus, CorridorCosts

from ...common.grid_functions import on_line


class CustomPriorityQueue:
    """Custom Priority Queue to allow for keeping the insertion order consistent."""

    def __init__(self):
        self.q = PriorityQueue()
        self.counter = 0

    def put(self, priority, item):
        count = self.counter
        self.counter += 1
        self.q.put((priority, count, item))

    def get(self):
        return self.q.get()[2]

    def empty(self):
        return self.q.empty()

    @property
    def queue(self):
        return self.q.queue


class CorridorBuilder:
    def __init__(self, maze):
        self.maze = maze
        self.rows = self.maze.rows
        self.cols = self.maze.cols

    ########## Prim's algorithm logic ########################
    def generate_corridor_prim(self) -> np.ndarray:
        """Generate corridors using an adaptation of Prim's algorithm.

        Builds the corridor around the rooms, without touching the rooms, so that
        the access points of the rooms need to be connected to the corridors.
        """
        grid = np.full((self.rows, self.cols), WALL)
        directions = [(2, 0), (-2, 0), (0, 2), (0, -2)]

        # Initialize the starting position
        current_position = self.maze.start_position
        grid[current_position] = PATH

        walls = []

        for d in directions:
            pos_valid, move_status = self.is_valid_next_position(
                grid, current_position, d
            )
            if pos_valid or move_status in [
                CorridorMoveStatus.ROOM_BOUNDARY,
                CorridorMoveStatus.MAZE_BOUNDARY,
            ]:
                walls.append((current_position, d))

        while walls:
            # Randomly select a wall from the list
            current_position, direction = self.maze.py_random.choice(walls)

            next_pos = (
                current_position[0] + direction[0],
                current_position[1] + direction[1],
            )

            intermediary_pos = (
                current_position[0] + direction[0] // 2,
                current_position[1] + direction[1] // 2,
            )

            is_valid, status = self.is_valid_next_position(
                grid, current_position, direction
            )

            if is_valid:
                grid[intermediary_pos[0], intermediary_pos[1]] = PATH
                grid[next_pos[0], next_pos[1]] = PATH

                for d in directions:
                    pos_valid, move_status = self.is_valid_next_position(
                        grid, next_pos, d
                    )
                    if pos_valid:
                        walls.append((next_pos, d))

            if not is_valid and self.maze.corridor_grid_connect_option:
                if (
                    0 <= intermediary_pos[0] < self.rows
                    and 0 <= intermediary_pos[1] < self.cols
                    and grid[intermediary_pos[0], intermediary_pos[1]] == WALL
                    and not self.is_adjacent_to_room(intermediary_pos)
                ):
                    grid[intermediary_pos[0], intermediary_pos[1]] = PATH

            # Remove the wall from the list
            walls.remove((current_position, direction))

        return grid

    def is_out_of_bounds_maze(self, position: Tuple[int, int]) -> bool:
        """
        Check if the given position is out of bounds in the maze.

        Args:
            position (Tuple[int, int]): The position to check.

        Returns:
            bool: True if the position is out of bounds, False otherwise.
        """
        if (
            position[0] < 0
            or position[0] >= self.rows
            or position[1] < 0
            or position[1] >= self.cols
        ):
            return True
        return False

    def is_valid_next_position(
        self, grid: np.ndarray, position: Tuple[int, int], direction: Tuple[int, int]
    ) -> Tuple[bool, CorridorMoveStatus]:
        """
        Check if the next position is a valid move in the grid.

        Args:
            grid (np.ndarray): The grid representing the maze.
            position (Tuple[int, int]): The current position in the grid.
            direction (Tuple[int, int]): The direction of the next move.

        Returns:
            Tuple[bool, CorridorMoveStatus]: A tuple containing a boolean indicating
                whether the next position is a valid move and a CorridorMoveStatus
                indicating the status of the move.
        """
        new_pos = (position[0] + direction[0], position[1] + direction[1])

        # Detect if new position is out of maze bounds
        if self.is_out_of_bounds_maze(new_pos):
            return (False, CorridorMoveStatus.MAZE_BOUNDARY)

        # Detect if new position or its neighbors are a path (either corridor or room)
        if (
            grid[new_pos[0], new_pos[1]] == PATH
            or self.maze.room_inner_area_grid[new_pos[0], new_pos[1]] == PATH
        ):
            return (False, CorridorMoveStatus.INVALID)

        # Check for room padding
        if self.is_adjacent_to_room(new_pos):
            return (False, CorridorMoveStatus.ROOM_BOUNDARY)

        return (True, CorridorMoveStatus.VALID_MOVE)

    def is_adjacent_to_room(self, position: Tuple[int, int]) -> bool:
        """
        Check if a given position is adjacent to any room in the maze.

        Parameters:
            position (Tuple[int, int]): The position to check adjacency for.

        Returns:
            bool: True if the position is adjacent to a room, False otherwise.
        """
        for dx in [-1, 0, 1]:
            for dy in [-1, 0, 1]:
                if (
                    0 <= position[0] + dx < self.rows
                    and 0 <= position[1] + dy < self.cols
                    and self.maze.room_inner_area_grid[
                        position[0] + dx, position[1] + dy
                    ]
                    == PATH
                ):
                    return True
        return False

    def greedy_sort(
        self, access_points: List[Tuple[int, int]], start_position: Tuple[int, int]
    ) -> List[Tuple[int, int]]:
        """
        Sorts a list of access points based on their proximity to the start position.
        At each step, the closest access point is added to the sorted list, and becomes the start_position reference.

        Args:
            access_points (List[Tuple[int, int]]): A list of tuples representing the coordinates of the access points.
            start_position (Tuple[int, int]): A tuple representing the coordinates of the start position.

        Returns:
            List[Tuple[int, int]]: A sorted list of tuples representing the coordinates of the access points.

        """
        # Initialize the sorted list with the start_position
        sorted_points = [start_position]

        # Remaining points to be sorted
        remaining_points = access_points.copy()

        while remaining_points:
            current_point = sorted_points[-1]  # Get the last added point
            closest_point = min(
                remaining_points, key=lambda point: self.heuristic(current_point, point)
            )

            sorted_points.append(closest_point)
            remaining_points.remove(closest_point)

        # Return the sorted list without the start_position
        return sorted_points[1:]

    ####### Connect access points to paths ############

    def connect_rooms_to_paths(self):
        """Connect the rooms to the corridors via the shortest path."""

        for room in self.maze.rooms:
            for access_point in room.access_points:
                global_access_point = (
                    room.global_position[0] + access_point[0],
                    room.global_position[1] + access_point[1],
                )
                connection = self.bfs_to_find_closest_path(global_access_point)
                if not connection:
                    # If BFS failed to find a connection, try growing a path
                    connection = self.grow_path_from(global_access_point)

                # If we found a connection, draw a path to it
                if connection:
                    self.plot_path_from_to(global_access_point, connection)

    def bfs_to_find_closest_path(
        self, access_point: Tuple[int, int]
    ) -> Optional[Tuple[int, int]]:
        directions = [(1, 0), (-1, 0), (0, 1), (0, -1)]

        visited = np.zeros((self.rows, self.cols), dtype=bool)
        queue = deque([access_point])

        # If the access point is already on a corridor path
        if self.maze.corridor_grid[access_point[0], access_point[1]] == PATH:
            return access_point

        while queue:
            current = queue.popleft()

            for d in directions:
                new_pos = (current[0] + d[0], current[1] + d[1])

                # Check boundaries
                if 0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols:
                    # If the new position is an existing path, we found a connection
                    if self.maze.corridor_grid[new_pos[0], new_pos[1]] == PATH:
                        return new_pos

                    # If not visited, not a room cell, not inside any room (excluding the access point), and not yet part of the corridor
                    if (
                        not visited[new_pos[0], new_pos[1]]
                        and not self.maze.is_inside_any_room(
                            new_pos, exception=access_point
                        )
                        and self.maze.corridor_grid[new_pos[0], new_pos[1]] == WALL
                    ):
                        visited[new_pos[0], new_pos[1]] = True
                        queue.append(new_pos)

        return None

    def plot_path_from_to(self, p1: Tuple[int, int], p2: Tuple[int, int]) -> None:
        """
        Plot a path from point p1 to point p2 in the maze.

        Args:
            p1 (Tuple[int, int]): The starting point coordinates.
            p2 (Tuple[int, int]): The ending point coordinates.

        Returns:
            None
        """
        # Directly connect if the cells are adjacent and not diagonally so
        diff_x = abs(p1[0] - p2[0])
        diff_y = abs(p1[1] - p2[1])
        if diff_x <= 1 and diff_y <= 1 and not (diff_x == 1 and diff_y == 1):
            self.maze.corridor_grid[p2[0], p2[1]] = PATH
            return

        while p1 != p2:
            # First adjust horizontally
            if p1[1] != p2[1]:
                p1 = (p1[0], p1[1] + np.sign(p2[1] - p1[1]))
            # Then adjust vertically
            else:
                p1 = (p1[0] + np.sign(p2[0] - p1[0]), p1[1])

            if not self.maze.is_inside_any_room(p1, exception=None):
                self.maze.corridor_grid[p1[0], p1[1]] = PATH

    ####### Grow path logic #########
    def grow_path_from(
        self, start_pos: Tuple[int, int], max_attempts: int = 1000
    ) -> Tuple[int, int]:
        """
        Generates a path from a given starting position within the maze.

        Args:
            start_pos (Tuple[int, int]): The starting position of the path.
            max_attempts (int, optional): The maximum number of attempts to generate the path. Defaults to 1000.

        Returns:
            Tuple[int, int]: The position of the generated path, or None if no path was found.
        """

        current_pos = start_pos
        attempt = 0
        directions = [(0, 1), (0, -1), (1, 0), (-1, 0)]

        visited = set()
        visited.add(start_pos)

        # Get a list of all corridor positions
        corridor_positions = np.argwhere(self.maze.corridor_grid == PATH)

        while attempt < max_attempts:
            potential_directions = []

            # Get the closest corridor position to the current position
            distances = [
                np.linalg.norm(np.array(current_pos) - np.array(corridor_pos))
                for corridor_pos in corridor_positions
            ]
            closest_corridor_pos = corridor_positions[np.argmin(distances)]

            for d in directions:
                new_pos = (current_pos[0] + d[0], current_pos[1] + d[1])

                if new_pos in visited:
                    continue

                # Check boundaries
                if not (0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols):
                    continue  # Go to the next direction if this is out of bounds

                # If the new position is an existing path, validate and then return
                if self.maze.corridor_grid[new_pos[0], new_pos[1]] == PATH:
                    diff_x = new_pos[0] - current_pos[0]
                    diff_y = new_pos[1] - current_pos[1]

                    # If this is not diagonal, then validate
                    if not (abs(diff_x) == 1 and abs(diff_y) == 1):
                        if not self.is_line_segment_intersecting_room(
                            start_pos, new_pos
                        ):
                            return new_pos
                    continue  # Go to the next direction

                # If not inside any room and not yet part of the corridor
                if (
                    not self.maze.is_inside_any_room(new_pos)
                    and self.maze.corridor_grid[new_pos[0], new_pos[1]] == WALL
                ):
                    potential_directions.append(
                        (
                            new_pos,
                            self.direction_cost(current_pos, d, closest_corridor_pos),
                        )
                    )

            # Sort the potential directions by their cost
            potential_directions.sort(key=lambda x: x[1])

            if not potential_directions:
                break

            # Select the best direction
            next_pos = potential_directions[0][0]
            visited.add(next_pos)

            # Update the grid
            self.maze.corridor_grid[next_pos[0], next_pos[1]] = PATH
            current_pos = next_pos
            attempt += 1

        # If we reached here, we didn't find a corridor
        return None

    def direction_cost(
        self,
        current_pos: Tuple[int, int],
        direction: Tuple[int, int],
        target_pos: Tuple[int, int],
    ) -> int:
        """
        Calculates the cost of moving in a given direction from the current position to the target position.

        Parameters:
            current_pos (Tuple[int, int]): The current position on the grid, represented as a tuple of integers (x, y).
            direction (Tuple[int, int]): The direction in which to move, represented as a tuple of integers (dx, dy).
            target_pos (Tuple[int, int]): The target position on the grid, represented as a tuple of integers (x, y).

        Returns:
            int: The cost of moving in the given direction from the current position to the target position. The cost is calculated based on the Manhattan distance between the new position and the target position, and is penalized if the new position is close to a room.
        """
        new_pos = (current_pos[0] + direction[0], current_pos[1] + direction[1])
        base_cost = abs(new_pos[0] - target_pos[0]) + abs(new_pos[1] - target_pos[1])

        # Penalize if the new position is close to a room.
        if self.maze.is_inside_any_room(new_pos):
            base_cost += 10

        return base_cost

    def is_line_segment_intersecting_room(
        self, p1: Tuple[int, int], p2: Tuple[int, int]
    ) -> bool:
        """
        Checks if a line segment defined by two points intersects with any room in the maze.

        Parameters:
            p1 (Tuple[int, int]): The first point of the line segment.
            p2 (Tuple[int, int]): The second point of the line segment.

        Returns:
            bool: True if the line segment intersects with any room, False otherwise.
        """

        for room in self.maze.rooms:
            for i in range(room.rows):
                for j in range(room.cols):
                    if room.grid[i][j] == PATH:
                        global_cell_pos = (
                            room.global_position[0] + i,
                            room.global_position[1] + j,
                        )

                        # Skip if it's the starting point
                        if global_cell_pos == p1:
                            continue

                        if on_line(p1, global_cell_pos, p2):
                            return True
        return False

    ###### Post Processing logic ######

    def post_process_maze(self):
        """
        Post-processes the maze by filling in paths between rooms and along the border.

        This function initializes the global room mask, collects border cells, shuffles them,
        and then attempts to fill in paths between the cells using the "_attempt_fill_path"
        method. It also collects the perimeter cells of all rooms, shuffles them, and
        attempts to fill in paths between these perimeter cells and the global room mask.

        Parameters:
            None

        Returns:
            None
        """
        # Initialize the global room mask
        global_room_mask = self.maze.generate_global_room_mask()

        # Collect border cells
        border_cells = []

        for i in range(self.rows):
            for j in [0, self.cols - 1]:  # Left and right borders
                border_cells.append((i, j))

        for j in range(self.cols):
            for i in [0, self.rows - 1]:  # Top and bottom borders
                border_cells.append((i, j))

        self.maze.py_random.shuffle(border_cells)  # Shuffle border cells

        for cell in border_cells:
            self._attempt_fill_path(cell, global_room_mask)

        # Collect the perimeter cells of all rooms
        room_perimeter_cells = []

        for room in self.maze.rooms:
            perimeter_cells = room.get_perimeter_cells(padding=2)
            for cell in perimeter_cells:
                # Translate to global maze coordinates
                global_pos = (
                    cell[0] + room.global_position[0],
                    cell[1] + room.global_position[1],
                )
                room_perimeter_cells.append(global_pos)

        self.maze.py_random.shuffle(room_perimeter_cells)  # Shuffle perimeter cells

        for cell in room_perimeter_cells:
            self._attempt_fill_path(cell, global_room_mask)

    def _attempt_fill_path(
        self, pos: Tuple[int, int], global_room_mask: np.ndarray
    ) -> None:
        """
        Attempts to fill a path at a given position in the maze.

        Args:
            pos (Tuple[int, int]): The position in the maze where the path is to be filled.
            global_room_mask (np.ndarray): The global room mask indicating the presence of rooms.

        Returns:
            None: This function does not return anything.
        """
        if pos[0] < 0 or pos[0] >= self.rows or pos[1] < 0 or pos[1] >= self.cols:
            return

        adjacent_dirs = [(0, 1), (0, -1), (1, 0), (-1, 0)]
        all_directions = adjacent_dirs + [(1, 1), (-1, 1), (-1, -1), (1, -1)]

        path_count = sum(
            [
                1
                for dx, dy in adjacent_dirs
                if 0 <= pos[0] + dx < self.rows
                and 0 <= pos[1] + dy < self.cols
                and self.maze.corridor_grid[pos[0] + dx, pos[1] + dy] == PATH
            ]
        )

        inside_or_adjacent_to_room = False
        for dx, dy in all_directions:
            new_pos = (pos[0] + dx, pos[1] + dy)
            if 0 <= new_pos[0] < self.rows and 0 <= new_pos[1] < self.cols:
                if global_room_mask[new_pos[0], new_pos[1]]:
                    inside_or_adjacent_to_room = True
                    break

        if not inside_or_adjacent_to_room and path_count == 1:
            self.maze.corridor_grid[pos[0], pos[1]] = PATH

    ####### Greedy best first search logic #########
    def generate_corridor_gbfs(self) -> np.ndarray:
        """
        Generates a corridor using the Greedy Best-First Search (GBFS) algorithm

        Returns:
            numpy.ndarray: The corridor grid with marked paths.
        """

        # Initialize a grid filled with walls
        corridor_grid = np.full((self.rows, self.cols), WALL)
        current_endpoint = self.maze.start_position

        access_points = self.get_global_access_points(current_endpoint)

        cost_grid = self.make_cost_grid(access_points, corridor_grid)

        for access_point in access_points:
            path = self.gbfs(current_endpoint, access_point, cost_grid)
            for point in path:
                corridor_grid[point[0], point[1]] = PATH
            current_endpoint = access_point

            # Recompute grid with the newly generated coridors
            cost_grid = self.make_cost_grid(access_points, corridor_grid)

        return corridor_grid

    def gbfs(
        self, start: Tuple[int, int], goal: Tuple[int, int], cost_grid: np.ndarray
    ) -> List[Tuple[int, int]]:
        """
        Performs the Greedy Best-First Search algorithm to find the path from the start point to the goal point in a grid.

        Parameters:
            start (Tuple[int, int]): The coordinates of the start point.
            goal (Tuple[int, int]): The coordinates of the goal point.
            cost_grid (np.ndarray): The cost grid representing the obstacles and costs of each cell.

        Returns:
            List[Tuple[int, int]]: The path from the start point to the goal point as a list of coordinates.
        """
        frontier = CustomPriorityQueue()
        frontier.put(0, start)

        came_from = {}
        came_from[start] = None

        while not frontier.empty():
            current = frontier.get()

            if current == goal:
                break

            directions = [(1, 0), (-1, 0), (0, 1), (0, -1)]
            self.maze.py_random.shuffle(directions)

            for dx, dy in directions:
                next_node = (current[0] + dx, current[1] + dy)

                if self.is_out_of_bounds_maze(next_node):
                    continue

                if next_node not in came_from:
                    priority = self.heuristic(goal, next_node, cost_grid)
                    frontier.put(priority, next_node)
                    came_from[next_node] = current

        path = self.reconstruct_path(came_from, start, goal)

        return path

    ####### A* direct corridor generation logic #########
    def generate_corridor_a_star(self) -> np.ndarray:
        """
        Generates a corridor using the A* algorithm.

        Returns:
            np.ndarray: The generated corridor grid with walls and paths.
        """

        # Initialize a grid filled with walls
        corridor_grid = np.full((self.rows, self.cols), WALL)
        current_endpoint = self.maze.start_position

        access_points = self.get_global_access_points(current_endpoint)

        cost_grid = self.make_cost_grid(access_points, corridor_grid)

        for access_point in access_points:
            path = self.a_star_path(current_endpoint, access_point, cost_grid)
            for point in path:
                corridor_grid[point[0], point[1]] = PATH
            current_endpoint = access_point

            # Recompute grid with the newly generated coridors
            cost_grid = self.make_cost_grid(access_points, corridor_grid)

        return corridor_grid

    def a_star_path(
        self, start: Tuple[int, int], goal: Tuple[int, int], cost_grid: np.ndarray
    ) -> List[Tuple[int, int]]:
        """
        Finds the shortest path from the start coordinate to the goal coordinate using the A* algorithm.

        Args:
            start (Tuple[int, int]): The starting coordinate.
            goal (Tuple[int, int]): The goal coordinate.
            cost_grid (np.ndarray): The grid containing the cost of each cell.

        Returns:
            List[Tuple[int, int]]: The shortest path from the start to the goal.
        """
        frontier = CustomPriorityQueue()
        frontier.put(0, start)

        came_from = {}
        came_from[start] = None
        cost_so_far = {}
        cost_so_far[start] = 0

        visited = set()
        visited.add(start)

        while not frontier.empty():
            current = frontier.get()

            if current == goal:
                break

            for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                next_node = (current[0] + dx, current[1] + dy)

                if next_node in visited:
                    continue

                if self.is_out_of_bounds_maze(next_node):
                    continue

                cell_cost = cost_grid[next_node[0], next_node[1]]
                new_cost = cost_so_far[current] + cell_cost
                if next_node not in cost_so_far or new_cost < cost_so_far[next_node]:
                    cost_so_far[next_node] = new_cost
                    priority = new_cost + self.heuristic(goal, next_node)
                    frontier.put(priority, next_node)
                    came_from[next_node] = current

        path = self.reconstruct_path(came_from, start, goal)

        return path

    ### Common astar and gbfs helper functions ###
    def get_global_access_points(
        self, current_endpoint: Tuple[int, int]
    ) -> List[Tuple[int, int]]:
        """
        Retrieves the access points in the rooms of the maze in global coordinates.
        Sorts the access points by heuristic to get even more direct paths.

        Parameters:
            current_endpoint (Tuple[int, int]): The current endpoint in the maze.

        Returns:
            List[Tuple[int, int]: A list of global access points in the maze.
        """

        access_points = []
        for room in self.maze.rooms:
            access_points_global = [
                (x + room.global_position[0], y + room.global_position[1])
                for x, y in room.access_points
            ]
            access_points.extend(access_points_global)

        # Sort the access points to get even more direct paths
        if self.maze.corridor_sort_access_points_option:
            access_points = self.greedy_sort(access_points, current_endpoint)
        else:
            self.maze.py_random.shuffle(access_points)

        return access_points

    def make_cost_grid(
        self, access_points: List[Tuple[int, int]], corridor_grid: np.ndarray
    ) -> np.ndarray:
        """
        Generates a cost grid based on the given access points and corridor grid.

        Parameters:
            access_points (List[Tuple[int, int]]): A list of tuples representing the coordinates of the access points.
            corridor_grid (np.ndarray): A numpy array representing the corridor grid.

        Returns:
            np.ndarray: A numpy array representing the cost grid.
        """
        cost_grid = np.full((self.rows, self.cols), CorridorCosts.BASE)

        global_room_mask = self.maze.generate_global_room_mask()
        cost_grid[global_room_mask] = CorridorCosts.ROOM_CELL

        for room in self.maze.rooms:
            global_pos = room.global_position
            perimeter_cells = room.get_perimeter_cells(padding=1)
            perimeter_cells_global = [
                (x + global_pos[0], y + global_pos[1]) for x, y in perimeter_cells
            ]
            for cell in perimeter_cells_global:
                if self.is_next_to_access_point(cell, access_points):
                    cost_grid[cell[0], cell[1]] = CorridorCosts.ADJACENT_ACCESS_POINT
                else:
                    cost_grid[cell[0], cell[1]] = CorridorCosts.ADJACENT_ROOM

            for access_point in access_points:
                cost_grid[access_point[0], access_point[1]] = CorridorCosts.ACCESS_POINT

        cost_grid[corridor_grid == PATH] = CorridorCosts.CORRIDOR

        return cost_grid

    def reconstruct_path(
        self,
        came_from: Dict[Tuple[int, int], Tuple[int, int]],
        start: Tuple[int, int],
        goal: Tuple[int, int],
    ) -> List[Tuple[int, int]]:
        """
        Reconstructs the path from the start to the goal using the came_from dictionary.

        Args:
            came_from (Dict[Tuple[int, int], Tuple[int, int]]): A dictionary that maps each cell to the cell it came from.
            start (Tuple[int, int]): The starting cell.
            goal (Tuple[int, int]): The goal cell.

        Returns:
            List[Tuple[int, int]]: The reconstructed path from the start to the goal.
        """
        current = goal
        path = []
        while current != start:
            path.append(current)
            current = came_from[current]
        path.append(start)
        path.reverse()

        return path

    def is_next_to_access_point(
        self,
        position: Tuple[int, int],
        access_points: Optional[List[Tuple[int, int]]] = None,
    ) -> bool:
        """
        Check if the given position is next to any access point.

        Parameters:
            position (Tuple[int, int]): The position to check.
            access_points (Optional[List[Tuple[int, int]]]): The list of access points to check against. If not provided, all access points in the maze will be considered.

        Returns:
            bool: True if the position is next to an access point, False otherwise.
        """
        if access_points is None:
            access_points = []
            for room in self.maze.rooms:
                access_points_global = [
                    (x + room.global_position[0], y + room.global_position[1])
                    for x, y in room.access_points
                ]
                access_points.extend(access_points_global)

        for access_point in access_points:
            for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
                if position == (access_point[0] + dx, access_point[1] + dy):
                    return True
        return False

    def heuristic(
        self,
        goal: Tuple[int, int],
        next_node: Tuple[int, int],
        cost_grid: Optional[np.ndarray] = None,
    ) -> Union[int, float]:
        """
        Calculate the heuristic value between the current node and the goal node.

        Args:
            goal (Tuple[int, int]): The coordinates of the goal node.
            next_node (Tuple[int, int]): The coordinates of the current node.
            cost_grid (Optional[np.ndarray], optional): A grid representing the cost of each node. Defaults to None.

        Returns:
            Union[int, float]: The calculated heuristic value.
        """
        heuristic = abs(goal[0] - next_node[0]) + abs(goal[1] - next_node[1])
        if cost_grid is not None:
            heuristic += cost_grid[next_node[0], next_node[1]]
        return heuristic



import gymnasium as gym
from gymnasium import spaces
import numpy as np
from scipy.linalg import expm
import networkx as nx
from .visualize import draw_network_graph, plot_centralities_sorted
from .graph_utils import (
    compute_laplacian,
    compute_eigenvector_centrality,
)


class NetworkGraph(gym.Env):
    metadata = {"render.modes": ["human", "matplotlib"]}

    def __init__(
        self,
        num_agents=10,
        connection_prob_range=(0.1, 0.3),
        max_u=0.1,
        desired_opinion=1.0,
        tau=1.0,
        initial_opinion_range=(0.0, 1.0),
        initial_opinions=None,
        control_resistance=None,
        connectivity_matrix=None,
        graph_connection_distribution="uniform",
        graph_connection_params=None,
        use_weighted_edges=False,
        weight_range=(0.1, 1.0),
        bidirectional_prob: float = 0.5,  # 1.0 = fully undirected, 0.0 = fully directed
        budget=None,
        max_steps=100,
        opinion_end_tolerance=0.01,
        control_beta=0.4,
        normalize_reward=False,
        terminal_reward=0.0,
        seed=None,
    ):
        super(NetworkGraph, self).__init__()

        self.seed = seed
        self._rng = np.random.default_rng(seed)
        
        if connectivity_matrix is not None:
            self.num_agents = connectivity_matrix.shape[0]
        else:
            self.num_agents = num_agents
            
        self.connection_prob_range = connection_prob_range
        self.use_weighted_edges = use_weighted_edges
        self.weight_range = weight_range

        if np.isscalar(max_u):
            self.max_u = np.full(self.num_agents, max_u, dtype=np.float32)
        else:
            assert len(max_u) == self.num_agents, "max_u vector must match number of agents"
            self.max_u = np.array(max_u, dtype=np.float32)

        self.desired_opinion = desired_opinion
        self.tau = tau
        self.initial_opinion_range = initial_opinion_range
        self.initial_opinions = initial_opinions
        self.control_resistance = control_resistance
        self.bidirectional_prob = bidirectional_prob

        self.budget = budget
        self.max_steps = max_steps
        self.opinion_end_tolerance = opinion_end_tolerance
        self.control_beta = control_beta
        self.normalize_reward = normalize_reward
        self.terminal_reward = terminal_reward
        
        if connectivity_matrix is not None:
            self.connectivity_matrix = connectivity_matrix
            if self.use_weighted_edges:
                for i in range(self.num_agents):
                    for j in range(self.num_agents):
                        if self.connectivity_matrix[i, j] == 1:
                            weight = self._rng.uniform(*self.weight_range)
                            self.connectivity_matrix[i, j] = weight
        else:
            graph_connection_params = graph_connection_params or {}

            # Generate probability matrix if needed
            if graph_connection_distribution == "uniform":
                prob = self._rng.uniform(*connection_prob_range)
                probs = np.full((self.num_agents, self.num_agents), prob)
            elif graph_connection_distribution == "normal":
                mean = graph_connection_params.get("mean", 0.2)
                std = graph_connection_params.get("std", 0.05)
                probs = np.clip(
                    self._rng.normal(mean, std, (self.num_agents, self.num_agents)), 0, 1
                )
            elif graph_connection_distribution == "exponential":
                scale = graph_connection_params.get("scale", 0.2)
                probs = np.clip(
                    self._rng.exponential(scale, (self.num_agents, self.num_agents)), 0, 1
                )
            else:
                raise ValueError(f"Unknown graph_connection_distribution: {graph_connection_distribution}")

            # Build directed graph with probabilistic bidirectionality
            self.graph = nx.DiGraph()
            self.graph.add_nodes_from(range(self.num_agents))
            for i in range(self.num_agents):
                for j in range(self.num_agents):
                    if i == j:
                        continue
                    if self._rng.random() < probs[i, j]:
                        self.graph.add_edge(i, j)
                        if self._rng.random() < self.bidirectional_prob:
                            self.graph.add_edge(j, i)

            self.connectivity_matrix = nx.to_numpy_array(self.graph)

            # Ensure weak connectivity by forcing at least one connection if isolated
            G = nx.DiGraph(self.connectivity_matrix)
            for node in range(self.num_agents):
                if G.in_degree(node) + G.out_degree(node) == 0:
                    candidates = list(range(self.num_agents))
                    candidates.remove(node)
                    neighbor = self._rng.choice(candidates)
                    G.add_edge(node, neighbor)
            self.connectivity_matrix = nx.to_numpy_array(G)

        # Laplacian and centrality
        self.L = compute_laplacian(self.connectivity_matrix)
        self.centralities = compute_eigenvector_centrality(self.L)

        if not nx.is_connected(nx.from_numpy_array(self.connectivity_matrix)):
            print("Warning: The generated graph is not fully connected.")

        if initial_opinions is not None:
            self.opinions = np.array(initial_opinions)
        else:
            self.opinions = self._rng.uniform(*self.initial_opinion_range, size=self.num_agents)

        if control_resistance is not None:
            self.control_resistance = np.array(control_resistance)
        else:
            self.control_resistance = np.zeros(self.num_agents)

        self.action_space = spaces.Box(
            low=np.zeros(self.num_agents, dtype=np.float32),
            high=self.max_u,
            shape=(self.num_agents,),
            dtype=np.float32,
        )
        self.observation_space = spaces.Box(
            low=0, high=1, shape=(self.num_agents,), dtype=np.float32
        )

        self.current_step = 0
        self.total_spent = 0.0

    @property
    def state(self):
        """
        Getter for opinions, allowing access via self.state.
        """
        return self.opinions

    @state.setter
    def state(self, value):
        """
        Setter for opinions, allowing assignment via self.state.
        """
        self.opinions = np.array(value)
        
    def reset(self, randomize_opinions=False):
        """
        Reset the environment to its initial state.

        Args:
            randomize_opinions (bool): If True, ignore self.initial_opinions and use random initialization instead.
        """
        if randomize_opinions or self.initial_opinions is None:
            self.opinions = self._rng.uniform(*self.initial_opinion_range, size=self.num_agents)
            info = {"random_opinions": True}
        else:
            self.opinions = np.array(self.initial_opinions)
            info = {"random_opinions": False}

        self.current_step = 0
        self.total_spent = 0.0
        return self.opinions, info

    def compute_dynamics(self, current_state, control_action, step_duration):
        """
        Compute the new state of the network given the current state, control action, and step duration.

        Parameters:
            current_state (numpy.ndarray): The current state of the network.
            control_action (numpy.ndarray): The control action to apply to the network.
            step_duration (float): The duration of the step.

        Returns:
            numpy.ndarray: The new state of the network.
        """
        # Ensure step duration is positive
        if step_duration <= 0:
            raise ValueError("step_duration must be positive")

        effective_control = control_action * (1 - self.control_resistance)

        # Apply control to blend current opinions with the desired opinion
        opinions = (
            effective_control * self.desired_opinion
            + (1 - effective_control) * current_state
        )

        # Compute opinion propagation with the influence of the Laplacian
        expL_remaining = expm(-self.L * step_duration)
        propagated_opinions = expL_remaining @ opinions

        # Clip the resulting opinions to be within [0, 1]
        new_states = np.clip(propagated_opinions, 0, 1)
        return new_states

    def reward_function(self, x, u, d, beta, done: bool = False):
        raw_reward = -np.abs(d - x).sum() - beta * np.sum(u)

        if self.normalize_reward:
            normalized_reward = raw_reward / self.num_agents
            reward = float(normalized_reward)
        else:
            reward = float(raw_reward)

        # Add terminal bonus if we're done due to reaching target
        if done:
            reward += self.terminal_reward

        return reward

    def step(self, action, step_duration=None):
        """
        Execute one time step within the environment.

        Args:
            action (np.array): Control inputs for influencing the agents.
            step_duration (float, optional): Total duration over which the opinions are propagated.
                                            Defaults to self.tau.

        Returns:
            observation (np.array): Updated opinions.
            reward (float): Reward for the step.
            done (bool): Whether the episode has ended.
            truncated (bool): Whether the episode was truncated (e.g., due to reaching the max number of steps).
            info (dict): Additional information.
        """
        if step_duration is None:
            step_duration = self.tau

        original_action = np.array(action, copy=True)  
        clipped_action = np.clip(original_action, 0, self.max_u)

        next_opinions = self.compute_dynamics(self.opinions, clipped_action, step_duration)

        self.total_spent += np.sum(original_action)
        self.current_step += 1
        
        done = np.abs(np.mean(next_opinions) - self.desired_opinion) <= self.opinion_end_tolerance
        truncated = self.current_step >= self.max_steps

        terminal_success = done and not truncated

        reward = self.reward_function(
            self.opinions, original_action, self.desired_opinion, self.control_beta, done=terminal_success
        )
        
        self.opinions = next_opinions
        
        info = {
            "current_step": self.current_step,
            "total_spent": self.total_spent,
            "action_applied_raw": original_action, 
            "action_applied_clipped": clipped_action,  
            "terminal_reward_applied": terminal_success,
        }
        
        if self.budget is not None:
            info["remaining_budget"] = self.budget - self.total_spent

        return self.opinions, reward, done, truncated, info

    def render(self, mode="matplotlib"):
        """
        Render the environment.

        Args:
            mode (str): The mode to render with. Supported modes: "human", "matplotlib".
        """
        if mode == "matplotlib":
            draw_network_graph(self.connectivity_matrix, self.centralities)
        elif mode == "centralities":
            plot_centralities_sorted(self.centralities)
        elif mode == "human":
            print(
                f"Step: {self.current_step}, Opinions: {self.opinions}, Total Spent: {self.total_spent}"
            )
        else:
            raise NotImplementedError(f"Render mode '{mode}' is not supported.")

    def close(self):
        """
        Clean up the environment.
        """
        pass



import numpy as np
import warnings
from scipy.optimize import minimize
from scipy.linalg import eig


def normalize_adjacency_matrix(A):
    row_sums = A.sum(axis=1, keepdims=True)
    row_sums[row_sums == 0] = 1
    A_normalized = A / row_sums
    return A_normalized


# Compute Laplacian
def compute_laplacian(adjacency_matrix):
    degree_matrix = np.diag(np.sum(adjacency_matrix, axis=1))
    laplacian = degree_matrix - adjacency_matrix
    return laplacian


# Compute centrality from Laplacian
def compute_eigenvector_centrality(L):
    eigenvalues, eigenvectors = eig(L, left=True, right=False)

    # This version seems to be an edge case that might not always work. Interesting
    # eigenvalues, eigenvectors = np.linalg.eig(L.T)
    min_eigenvalue_index = np.argmin(np.abs(eigenvalues))
    eigv = np.real(eigenvectors[:, min_eigenvalue_index])
    eigv_normalized = np.abs(eigv) / np.sum(np.abs(eigv))
    return eigv_normalized


def process_multiple_components(components):
    """
    Processes multiple components and computes the centrality for each.

    Parameters:
    components (list of numpy.ndarray): List of adjacency matrices for each component.

    Returns:
    list of numpy.ndarray: Normalized eigenvector centralities for each component.
    """
    centralities = []
    for component in components:
        L_component = compute_laplacian(component)
        centrality = compute_eigenvector_centrality(L_component)
        centralities.append(centrality)
    return centralities


def compute_centrality_for_component(L):
    """
    Computes centrality for a single connected component.

    Parameters:
    L (numpy.ndarray): Laplacian matrix of the connected component.

    Returns:
    numpy.ndarray: The computed centrality vector for the component.
    """
    eigenvalues, eigenvectors = np.linalg.eig(L.T)

    # Find the index of the smallest eigenvalue (which should be close to zero)
    min_eigenvalue_index = np.argmin(np.abs(eigenvalues))

    # The corresponding eigenvector is used as the centrality vector
    centrality_vector = np.abs(eigenvectors[:, min_eigenvalue_index]).real

    # Normalize the centrality vector to sum to 1
    if np.sum(centrality_vector) > 0:
        centrality_vector /= np.sum(centrality_vector)

    return centrality_vector


def compute_centrality(L, adjacency_matrix):
    """
    Computes centrality for both connected and disconnected graphs by
    computing centralities for each connected component.

    Parameters:
    L (numpy.ndarray): Laplacian matrix of the graph.
    adjacency_matrix (numpy.ndarray): Adjacency matrix of the graph.

    Returns:
    numpy.ndarray: The computed centrality vector for the entire graph.
    """
    num_nodes = L.shape[0]
    centrality = np.zeros(num_nodes)
    visited = np.zeros(num_nodes, dtype=bool)

    # Identify connected components
    for i in range(num_nodes):
        if not visited[i]:
            # Find the nodes in the same connected component as node i
            component = np.where(
                np.linalg.matrix_power(adjacency_matrix + np.eye(num_nodes), num_nodes)[
                    i
                ]
                > 0
            )[0]
            component_L = L[np.ix_(component, component)]
            component_centrality = compute_centrality_for_component(component_L)

            # Check if the entire component is isolated (no incoming or outgoing edges)
            has_incoming_edges = np.any(adjacency_matrix[:, component], axis=0)
            is_isolated_component = not np.any(has_incoming_edges)

            if not is_isolated_component:
                # Assign centrality to non-isolated nodes
                centrality[component] = component_centrality
            visited[component] = True

    # Normalize the centrality vector for non-isolated components
    connected_centrality_sum = np.sum(centrality)
    if connected_centrality_sum > 0:
        centrality /= connected_centrality_sum

    return centrality


def get_weighted_adjacency_matrix(
    connectivity_matrix, desired_centrality, tolerance=0.0001
):
    num_edges = np.sum(connectivity_matrix)  # Number of edges (non-zero entries)

    if num_edges == 0:
        raise ValueError("The connectivity matrix has no edges.")

    initial_weights = np.ones(int(num_edges))  # Start with equal weights for all edges

    bounds = [(0, None)] * int(num_edges)  # Weights should be non-negative

    def objective(weights, connectivity_matrix, desired_centrality):
        weighted_adjacency_matrix = np.zeros_like(connectivity_matrix, dtype=float)
        weighted_adjacency_matrix[connectivity_matrix == 1] = weights
        L = compute_laplacian(weighted_adjacency_matrix)
        computed_centrality = compute_centrality(L, weighted_adjacency_matrix)
        return np.linalg.norm(computed_centrality - desired_centrality)

    result = minimize(
        objective,
        initial_weights,
        args=(connectivity_matrix, desired_centrality),
        method="L-BFGS-B",
        bounds=bounds,
    )

    optimal_weights = result.x
    weighted_adj_matrix = np.zeros_like(connectivity_matrix, dtype=float)

    # Assign weights to the adjacency matrix
    weighted_adj_matrix[connectivity_matrix == 1] = optimal_weights

    # Compute the Laplacian and resulting centrality from the optimized adjacency matrix
    L = compute_laplacian(weighted_adj_matrix)
    computed_centrality = compute_centrality(L, weighted_adj_matrix)

    # Check if the computed centrality is close to the desired centrality
    if not np.allclose(computed_centrality, desired_centrality, atol=tolerance):
        warnings.warn(
            f"Resulting centrality is not within the tolerance of {tolerance}. "
            f"Difference: {np.abs(computed_centrality - desired_centrality)}",
            UserWarning,
        )

    return weighted_adj_matrix, computed_centrality



import networkx as nx
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

sns.set_theme()


def draw_network_graph(adjacency_matrix, node_sizes):
    G = nx.DiGraph(adjacency_matrix)
    node_sizes = [5000 * centrality for centrality in node_sizes]

    # Use Spring layout with adjusted k parameter to prevent node overlap
    pos = nx.spring_layout(G, k=2 / (len(G) ** 0.5), iterations=50)

    nx.draw_networkx_nodes(
        G, pos, node_size=node_sizes, node_color="skyblue", alpha=0.8
    )
    nx.draw_networkx_labels(G, pos, font_size=12, font_weight="bold")
    nx.draw_networkx_edges(G, pos, edge_color="gray", arrowsize=20, arrowstyle="->")

    for i, j in G.edges():
        if G.has_edge(j, i):
            nx.draw_networkx_edges(
                G,
                pos,
                edgelist=[(i, j)],
                edge_color="gray",
                arrowsize=20,
                arrowstyle="<->",
            )

    plt.title("Network Graph")
    plt.show()


def plot_centralities_sorted(eigv):
    indices = np.arange(len(eigv))
    sorted_indices = np.argsort(eigv)[::-1]
    sorted_centralities = eigv[sorted_indices]
    sorted_labels = [str(i) for i in indices[sorted_indices]]

    # Create a Seaborn barplot
    plt.figure(figsize=(10, 6))
    sns.barplot(x=sorted_labels, y=sorted_centralities, palette="Blues_d")

    # Set labels and title
    plt.xlabel("Eigenvalue Index")
    plt.ylabel("Centrality")
    plt.title("Centralities Sorted in Decreasing Order")
    plt.xticks(rotation=45)
    plt.show()



import gymnasium as gym
import numpy as np
from typing import Tuple


class GamblersProblem(gym.Env):
    """
    Gambler's Problem environment.
    """

    def __init__(
        self,
        goal_amount: int = 100,
        win_probability: float = 0.4,
        start_capital: int = 1,
    ):
        """
        Initialize the environment.

        Args:
            goal_amount (int): The goal capital amount to be achieved.
            win_probability (float): Probability of winning a bet.
        """
        self.goal_amount = goal_amount
        self.win_probability = win_probability
        self.start_capital = start_capital

        # Action space: bet an amount from 0 to the current capital
        # Observation space: current capital amount
        self.action_space = gym.spaces.Discrete(goal_amount + 1)
        self.observation_space = gym.spaces.Discrete(goal_amount + 1)

        self.state = None
        self.reset()

    def step(self, action: int) -> Tuple[int, float, bool, dict]:
        """
        Execute one time step within the environment.

        Args:
            action (int): Amount of capital to bet.

        Returns:
            Tuple containing the new state, reward, done flag, and additional info.
        """
        assert self.action_space.contains(action), "Invalid action"

        done = False
        reward = 0

        # Win case
        if np.random.rand() < self.win_probability:
            self.state += action
            if self.state >= self.goal_amount:
                self.state = self.goal_amount
                reward = 1
                done = True
        else:
            # Lose case
            self.state -= action
            if self.state <= 0:
                self.state = 0
                done = True

        truncated = False
        return self.state, reward, done, truncated, {}

    def reset(self) -> int:
        """
        Reset the environment to an initial state.

        Returns:
            The initial state and additional info.
        """
        self.state = self.start_capital
        return self.state, {}

    def render(self, mode="human") -> str:
        """
        Render the environment.
        """
        if mode == "human":
            print(f"Current Capital: ${self.state}")
        elif mode == "ansi":
            return f"Current Capital: ${self.state}"
        else:
            raise NotImplementedError("Render mode not supported: " + mode)

    def build_mdp(self) -> dict:
        """
        Build the MDP representation for the Gambler's Problem.
        """
        t_r_dict = {}

        for state in range(1, self.goal_amount):  # States from 1 to goal_amount - 1
            for action in range(
                1, min(state, self.goal_amount - state) + 1
            ):  # Possible bets
                # Winning scenario
                win_state = min(state + action, self.goal_amount)
                win_reward = 1 if win_state == self.goal_amount else 0
                win_prob = self.win_probability

                # Losing scenario
                lose_state = max(state - action, 0)
                lose_reward = 0
                lose_prob = 1 - self.win_probability

                t_r_dict[(state, action)] = {
                    "win": (
                        win_state,
                        win_reward,
                        win_state == self.goal_amount,
                        win_prob,
                    ),
                    "loss": (lose_state, lose_reward, lose_state == 0, lose_prob),
                }

        return t_r_dict



import numpy as np
import random
from collections import deque
from typing import List, Tuple, Union, Optional
from enum import Enum

from .room import RoomFactory, Room
from .corridor import CorridorBuilder
from ..constants import WALL, PATH, TARGET, CorridorMoveStatus

from ...common.grid_functions import on_line


class MazeFactory:
    def __init__(
        self,
        rows: int,
        cols: int,
        nr_desired_rooms: Optional[int] = None,
        nr_desired_rooms_range: Tuple[int, int] = (1, 8),
        global_room_ratio: Optional[float] = None,
        global_room_ratio_range: Tuple[float, float] = (0.1, 0.8),
        access_points_per_room: Optional[int] = None,
        access_points_per_room_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        room_ratio: Optional[Union[int, float]] = None,
        room_ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        corridor_algorithm: Optional[str] = "random",
        corridor_grid_connect_option: Optional[Union[bool, str]] = "random",
        corridor_post_process_option: bool = True,
        corridor_sort_access_points_option: Optional[Union[bool, str]] = "random",
        seed: Optional[int] = None,
    ):
        """
        Maze builder to randomize the generated maze.

        Arguments num_rooms, global_room_ratio, access_points_per_room
        are used to fix specific values, otherwise the values are drawn from the minimum and maximum distibutions.

        room_types is used to determine what types of rooms are to be added in the maze, if None, the random seletion
        considers all the implemented room types.

        Args:
            rows (int): The number of rows in the maze.
            cols (int): The number of columns in the maze.
            nr_desired_rooms (int, optional): The desired number of rooms in the maze. Defaults to None.
            nr_desired_rooms_range (tuple, optional): The range of desired number of rooms. Defaults to (1, 8).
            global_room_ratio (float, optional): The global room ratio in the maze. Defaults to None.
            global_room_ratio_range (tuple, optional): The range of global room ratio. Defaults to (0.1, 0.8).
            access_points_per_room (int, optional): The number of access points in each room. Defaults to None.
            access_points_per_room_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added in the maze. Defaults to None.
            room_ratio (float, optional): The room ratio. Defaults to None.
            room_ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            corridor_algorithm (str, optional): The algorithm to use for generating the maze. Defaults to 'random'. Can be one of: 'prim', 'astar', 'gbfs' or 'random'. if 'random', randomly selects algorithm.
            corridor_grid_connect_option (Union[bool, str], optional): Option to decide the nature of corridor paths connectivity.
                - If `True`, corridors will be grid-connected.
                - If `False`, corridors will not be grid-connected.
                - If `"random"`, the choice will be made randomly.
                Defaults to False.
            corridor_post_process_option (bool, optional): Whether to post-process the maze. Defaults to True.
            corridor_sort_access_points_option (Union[bool, str], str, optional): Whether to sort the access points. Defaults to "random". If "random" will randomly select between True and False.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """

        self.rows = rows
        self.cols = cols

        self.nr_desired_rooms = nr_desired_rooms
        self.nr_desired_rooms_range = nr_desired_rooms_range

        self.global_room_ratio = global_room_ratio
        self.global_room_ratio_range = global_room_ratio_range

        self.access_points_per_room = access_points_per_room
        self.access_points_per_room_range = access_points_per_room_range

        self.room_types = room_types

        self.room_ratio = room_ratio
        self.room_ratio_range = room_ratio_range

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)
        self.py_random = random.Random(self.seed)
        self.np_random = np.random.RandomState(self.seed)

        self.corridor_algorithm = corridor_algorithm
        self._check_corridor_algorithm_option()

        self.corridor_grid_connect_option = corridor_grid_connect_option
        self._check_corridor_grid_connect_option()

        self.corridor_post_process_option = corridor_post_process_option

        self.corridor_sort_access_points_option = corridor_sort_access_points_option
        self._check_corridor_sort_access_points_option()

    def create_maze(self) -> "Maze":
        """
        Creates a maze based on the provided parameters.

        Parameters:
            None

        Returns:
            maze (Maze): The created maze object.

        Raises:
            ValueError: If the global room ratio is greater than 1 or if the global room ratio range is invalid.

        """
        # Decide the number of rooms
        if self.nr_desired_rooms:
            num_rooms = self.nr_desired_rooms
        else:
            num_rooms = self.py_random.randint(
                self.nr_desired_rooms_range[0], self.nr_desired_rooms_range[1]
            )

        # Decide the global room ratio
        if self.global_room_ratio:
            if self.global_room_ratio > 1:
                raise ValueError(
                    "Global room ratio must be less than 1, but got {self.global_room_ratio}."
                )
            global_room_ratio = self.global_room_ratio
        else:
            if (
                self.global_room_ratio_range[0] < 0
                or self.global_room_ratio_range[1] > 1
            ):
                raise ValueError(
                    f"Global room ratio range must be between 0 and 1, but got {self.global_room_ratio_range}."
                )
            global_room_ratio = self.py_random.uniform(
                self.global_room_ratio_range[0], self.global_room_ratio_range[1]
            )

        # Randomize or fix other parameters
        self._check_corridor_algorithm_option()
        if self.corridor_algorithm == "random":
            corridor_algorithm = self.py_random.choice(["prim", "astar", "gbfs"])
        else:
            corridor_algorithm = self.corridor_algorithm

        self._check_corridor_grid_connect_option()
        if self.corridor_grid_connect_option == "random":
            corridor_grid_connect_option = self.py_random.choice([True, False])
        else:
            corridor_grid_connect_option = self.corridor_grid_connect_option

        self._check_corridor_sort_access_points_option()
        if self.corridor_sort_access_points_option == "random":
            corridor_sort_access_points_option = self.py_random.choice([True, False])
        else:
            corridor_sort_access_points_option = self.corridor_sort_access_points_option

        maze_seed = self.seed + 1
        maze = Maze(
            rows=self.rows,
            cols=self.cols,
            nr_desired_rooms=num_rooms,
            global_room_ratio=global_room_ratio,
            access_points_per_room=self.access_points_per_room,
            access_points_per_room_range=self.access_points_per_room_range,
            room_types=self.room_types,
            room_ratio=self.room_ratio,
            room_ratio_range=self.room_ratio_range,
            corridor_algorithm=corridor_algorithm,
            corridor_grid_connect_option=corridor_grid_connect_option,
            corridor_post_process_option=self.corridor_post_process_option,
            corridor_sort_access_points_option=corridor_sort_access_points_option,
            seed=maze_seed,
        )
        return maze

    def _check_corridor_grid_connect_option(self):
        if self.corridor_grid_connect_option not in [True, False, "random"]:
            raise ValueError(
                f"Invalid value for corridor_grid_connect_option. Got {self.corridor_grid_connect_option}, expected [True, False, 'random']."
            )

    def _check_corridor_algorithm_option(self):
        if self.corridor_algorithm not in ["prim", "astar", "gbfs", "random"]:
            raise ValueError(
                f"Invalid value for corridor_algorithm. Got {self.corridor_algorithm}, expected ['prim', 'astar', 'gbfs', 'random']."
            )

    def _check_corridor_sort_access_points_option(self):
        if self.corridor_sort_access_points_option not in [True, False, "random"]:
            raise ValueError(
                f"Invalid value for corridor_sort_access_points_option. Got {self.corridor_sort_access_points_option}, expected [True, False, 'random']."
            )


class Maze:
    def __init__(
        self,
        rows: int,
        cols: int,
        layout: Optional[np.ndarray] = None,
        nr_desired_rooms: int = 3,
        global_room_ratio: float = 0.5,
        access_points_per_room: Optional[int] = None,
        access_points_per_room_range: Tuple[int, int] = (1, 4),
        room_types: Optional[List[str]] = None,
        room_ratio: Optional[Union[int, float]] = None,
        room_ratio_range: Tuple[Union[int, float], Union[int, float]] = (0.5, 1.5),
        corridor_algorithm: Optional[str] = "prim",
        corridor_grid_connect_option: Optional[bool] = False,
        corridor_post_process_option: Optional[bool] = True,
        corridor_sort_access_points_option: Optional[bool] = False,
        seed: Optional[int] = None,
    ) -> None:
        """Construct a Maze object representing a maze layout.

        Args:
            rows (int): The number of rows in the maze.
            cols (int): The number of columns in the maze.
            layout (np.ndarray, optional): The layout of the maze. Defaults to None. Leads to overriding the generation behavior.
            nr_desired_rooms (int, optional): The desired number of rooms in the maze. Defaults to 3.
            global_room_ratio (float, optional): The global room ratio in the maze. Defaults to 0.5.
            access_points_per_room (int, optional): The number of access points in each room. Defaults to None.
            access_points_per_room_range (tuple, optional): The range of access points per room. Defaults to (1, 4).
            room_types (list, optional): The types of rooms to be added in the maze. Defaults to None.
            room_ratio (float, optional): The room ratio. Defaults to None.
            room_ratio_range (tuple, optional): The range of room ratio. Defaults to (0.5, 1.5).
            algorithm (str, optional): The algorithm to use for generating the maze. Defaults to 'prim'. Can be one of: 'prim', 'astar'.
            corridor_grid_connect_option (bool, optional): Option to decide the nature of corridor paths connectivity.
                - If `True`, corridors will be grid-connected.
                - If `False`, corridors will not be grid-connected.
                Defaults to False. Only affects 'prim' algorithm.
            corridor_grid_connect_option (bool, optional): Whether to post-process the maze. Defaults to True. Only affects 'prim' algorithm.
            corridor_sort_access_points_option (bool, optional): Whether to sort the access points in corridor generation. Defaults to False.
            seed (int, optional): The seed to use for generating random numbers. Defaults to None.
        """
        if layout is None and (rows < 10 or cols < 10):
            raise ValueError(
                "Maze dimensions must be at least 10x10, otherwise the maze generation logic risks breaking."
            )

        self.rows = rows
        self.cols = cols

        self.seed = seed
        if self.seed is None:
            self.seed = random.randint(0, 1e6)
        self.py_random = random.Random(self.seed)

        self.grid = np.ones((rows, cols), dtype=int) * WALL
        self.room_grid = np.ones((rows, cols), dtype=int) * WALL
        self.room_inner_area_grid = np.zeros((rows, cols), dtype=int)
        self.corridor_grid = np.ones((rows, cols), dtype=int) * WALL

        self.rooms = []
        self.nr_desired_rooms = nr_desired_rooms

        # in some cases we won't be able to place all rooms we want so track them
        self.nr_placed_rooms = 0
        self.global_room_ratio = global_room_ratio

        self.start_position = None
        self.target_position = None

        # Make Room Factory
        room_factory_seed = self.seed + 1
        self.room_factory = RoomFactory(
            access_points_nr=access_points_per_room,
            access_points_range=access_points_per_room_range,
            room_types=room_types,
            ratio=room_ratio,
            ratio_range=room_ratio_range,
            seed=room_factory_seed,
        )

        # Corridor generationg settings
        self.corridor_algorithm = corridor_algorithm
        self.corridor_grid_connect_option = corridor_grid_connect_option
        self.corridor_post_process_option = corridor_post_process_option
        self.corridor_sort_access_points_option = corridor_sort_access_points_option

        # Make Corridor Builder
        self.corridor_builder = CorridorBuilder(self)

        if layout is None:
            self._build_maze()
        else:
            self.grid = layout

    def _build_maze(self):
        """
        Builds the maze by making rooms and corridors.

        Parameters:
            None

        Returns:
            None
        """
        self.make_rooms()
        self.grid = np.where(self.room_grid == PATH, PATH, self.grid)

        self.make_corridors()
        self.grid = np.where(self.corridor_grid == PATH, PATH, self.grid)

    def make_corridors(self) -> None:
        """
        Generates the corridors for the maze.

        This function generates the corridors for the maze by using the specified corridor algorithm.
        It first places the start and end positions in the maze using the `place_start_end_positions` method.

        If the corridor algorithm is set to "prim", it generates the corridors using the Prim's algorithm
        implemented in the `generate_corridor_prim` method of the `corridor_builder` object. After generating
        the corridors, it connects the rooms to the paths using the `connect_rooms_to_paths` method of the
        `corridor_builder` object. If the `corridor_post_process_option` is set, it also applies post-processing
        to the maze using the `post_process_maze` method of the `corridor_builder` object.

        If the corridor algorithm is set to "astar", it generates the corridors using the A* algorithm
        implemented in the `generate_corridor_a_star` method of the `corridor_builder` object.

        If the corridor algorithm is set to "gbfs", it generates the corridors using the Greedy Best-First Search
        algorithm implemented in the `generate_corridor_gbfs` method of the `corridor_builder` object.
        """
        self.place_start_end_positions()

        if self.corridor_algorithm == "prim":
            self.corridor_grid = self.corridor_builder.generate_corridor_prim()
            self.corridor_builder.connect_rooms_to_paths()
            if self.corridor_post_process_option:
                self.corridor_builder.post_process_maze()

        elif self.corridor_algorithm == "astar":
            self.corridor_grid = self.corridor_builder.generate_corridor_a_star()

        elif self.corridor_algorithm == "gbfs":
            self.corridor_grid = self.corridor_builder.generate_corridor_gbfs()

    ##### Room generation and placement #####
    def levy_flight_place(self, room: Room) -> bool:
        """
        Places a room in the maze using a Levy flight algorithm.

        Parameters:
            room (Room): The room to be placed in the maze.

        Returns:
            bool: True if the room is successfully placed, False otherwise.
        """

        max_attempts = 100
        attempt = 0

        margin_padding = 2

        # define range from 2 to len - (minsize + 2) because we don't want to place rooms too close to edge of maze
        row_padding = room.rows + margin_padding
        col_padding = room.cols + margin_padding

        bottom_right_possible_row_coord = self.grid.shape[0] - row_padding
        bottom_right_possible_col_coord = self.grid.shape[1] - col_padding

        if (
            bottom_right_possible_row_coord < margin_padding
            or bottom_right_possible_col_coord < margin_padding
        ):
            # This proposed room is too big
            return False

        #### Attempt to find a viable position ####
        position = (
            self.py_random.randint(margin_padding, bottom_right_possible_row_coord),
            self.py_random.randint(margin_padding, bottom_right_possible_col_coord),
        )
        directions = [(0, 1), (0, -1), (1, 0), (-1, 0)]

        while attempt < max_attempts:
            if self.is_valid_room_position(room, position[0], position[1]):
                self.rooms.append(room)
                room.global_position = position
                self.materialize_room(room, position[0], position[1])
                return True
            else:
                step_size = self.levy_step_size()
                direction = self.py_random.choice(directions)
                position = (
                    position[0] + direction[0] * step_size,
                    position[1] + direction[1] * step_size,
                )
                attempt += 1

        return False

    def make_rooms(self) -> None:
        """
        Generates rooms until a certain condition is met.

        This function generates rooms until one of the following conditions is met:
        - The number of placed rooms is equal to the desired number of rooms.
        - The total area covered by the rooms is equal to the target area.
        - The maximum number of attempts is reached.

        Parameters:
        - None

        Returns:
        - None
        """
        total_room_path_area_covered = 0
        target_path_area = self.global_room_ratio * self.grid.size
        start_avg_room_path_area = target_path_area / self.nr_desired_rooms
        desired_room_path_area = start_avg_room_path_area

        # Because of the desired_room_path_area = max(9, desired_room_path_area),
        # we need to also check for a maximum attempt counter
        attempts = 0
        max_attempts = 100

        room_sizes = []
        while (
            self.nr_placed_rooms < self.nr_desired_rooms
            and total_room_path_area_covered < target_path_area
            and attempts < max_attempts
        ):
            # In some cases when the global ratio is small and we want many rooms,
            # the start_avg is small from the very start so we need to enforce minimum area
            desired_room_path_area = max(9, desired_room_path_area)

            room = self.room_factory.create_room(desired_area=desired_room_path_area)
            room_sizes.append((room.rows, room.cols))
            if self.levy_flight_place(room):
                total_room_path_area_covered += room.rows * room.cols
                self.nr_placed_rooms += 1

            # Each time a room is added, or a fail happens, decrease the desired area of the next room trials
            desired_room_path_area = int(0.9 * desired_room_path_area)
            attempts += 1

    def is_valid_room_position(
        self, room: Room, start_row: int, start_col: int
    ) -> bool:
        """
        Checks if a room position is valid within the room grid.

        Args:
            room (Room): The room to be placed.
            start_row (int): The starting row index of the room position.
            start_col (int): The starting column index of the room position.

        Returns:
            bool: True if the room position is valid, False otherwise.
        """
        end_row, end_col = start_row + room.rows, start_col + room.cols

        # Extract sub-grid and check boundaries
        if (
            start_row < 1
            or start_col < 1
            or end_row > self.room_grid.shape[0] - 1
            or end_col > self.room_grid.shape[1] - 1
        ):
            return False

        # Basically add a padding of 1
        sub_grid = self.room_inner_area_grid[
            start_row - 1 : end_row + 1, start_col - 1 : end_col + 1
        ]

        # Check if placing the room would overlap with another room or not maintain separation
        return np.sum(sub_grid == PATH) == 0

    def materialize_room(self, room, start_row: int, start_col: int):
        """room_inner_area_grid -> used to find spacing and build corridors
        (this was created in special for the cases where we have rooms with hollow areas)

        room_grid -> actual PATHs that belong to rooms

        Args:
            room (Room): room object to be placed in the maze
            start_row (int): global row position of the room
            start_col (int): global row position of the room
        """
        self.room_inner_area_grid[
            start_row : start_row + room.rows, start_col : start_col + room.cols
        ] = room.generate_inner_area_mask()
        self.room_grid[
            start_row : start_row + room.rows, start_col : start_col + room.cols
        ] = room.grid

    def levy_step_size(self):
        """
        Generate the step size for the Levy flight algorithm.

        Returns:
            int: The step size for the Levy flight algorithm.

        Algorithm:
            1. Generate a random number between 0 and 1.
            2. Add 0.001 to the random number.
            3. Take the square root of the sum.
            4. Take the reciprocal of the square root.
            5. Convert the result to an integer.

        This function uses the inverse square distribution to calculate the step size for the Levy flight algorithm.
        The step size is a measure of the distance covered by the algorithm in each iteration.
        """
        r = self.py_random.random()  # r is between 0 and 1
        return int(
            1 / ((r + 0.001) ** 0.5)
        )  # This will give us a step length L according to inverse square distribution.

    ##### Corridor generation #####

    def choose_start_position(self) -> Tuple[int, int]:
        """
        Choose a random starting position for the agent.

        Returns:
            A tuple containing the row and column indices of the starting position.
        """
        side = self.py_random.choice(["top", "bottom", "left", "right"])

        if side == "top":
            return (0, self.py_random.randint(0, self.grid.shape[1] - 1))
        elif side == "bottom":
            return (
                self.grid.shape[0] - 1,
                self.py_random.randint(0, self.grid.shape[1] - 1),
            )
        elif side == "left":
            return (self.py_random.randint(0, self.grid.shape[0] - 1), 0)
        else:  # 'right'
            return (
                self.py_random.randint(0, self.grid.shape[0] - 1),
                self.grid.shape[1] - 1,
            )

    def choose_target_position(self) -> Tuple[int, int]:
        """
        Choose a target position in the grid.

        Returns:
            Tuple[int, int]: The row and column indices of the chosen target position.

        Raises:
            ValueError: If a valid target position could not be found in any room.
        """
        # Shuffle the rooms to ensure random selection without replacement
        shuffled_rooms = self.py_random.sample(self.rooms, len(self.rooms))

        # Preference to place the target in a non perimeter cell
        for random_room in shuffled_rooms:
            room_top_left_row, room_top_left_col = random_room.global_position

            # Create a list of all possible cell positions in the room
            non_perimeter_mask = random_room.get_non_perimeter_inner_cells()
            path_mask = random_room.grid == PATH
            combined_mask = non_perimeter_mask * path_mask
            candidates = np.argwhere(combined_mask == 1).tolist()

            # Shuffle these positions to randomly select without replacement
            self.py_random.shuffle(candidates)

            if len(candidates) > 0:
                row, col = candidates[0]
                return (room_top_left_row + row, room_top_left_col + col)

        # if no such cell could not be found, place target on a non access perimeter
        for random_room in shuffled_rooms:
            perimeter_cells = random_room.get_perimeter_cells()
            non_acces_perimeter_cells = [
                cell
                for cell in perimeter_cells
                if cell not in random_room.access_points
            ]
            if len(non_acces_perimeter_cells) > 0:
                self.py_random.shuffle(non_acces_perimeter_cells)
                return non_acces_perimeter_cells[0]

        raise ValueError(f"Could not find a valid target position in any room.")

    def place_start_end_positions(self) -> None:
        """
        Place the start and end positions of the game.

        This method is responsible for choosing the start and target positions for the game. It calls the "choose_start_position()" method to select the start position and the "choose_target_position()" method to select the target position. The start position and target position are stored in the "start_position" and "target_position" attributes respectively.

        Parameters:
        - None

        Returns:
        - None
        """
        self.start_position = self.choose_start_position()
        self.target_position = self.choose_target_position()

    def is_inside_any_room(
        self, pos: Tuple[int, int], exception: Tuple[int, int] = None
    ) -> bool:
        """
        Check if a given position is inside any room in the current environment.

        Parameters:
            pos (Tuple[int, int]): The position to check, represented as a tuple of integers (row, column).
            exception (Tuple[int, int], optional): A position that should be excluded from the check. Defaults to None.

        Returns:
            bool: True if the position is inside any room, False otherwise.
        """
        for room in self.rooms:
            room_mask = room.generate_inner_area_mask()

            # Convert the local room position to a global position
            global_row = pos[0] - room.global_position[0]
            global_col = pos[1] - room.global_position[1]

            # Check if the global position is within the bounds of the room's grid
            if (0 <= global_row < room_mask.shape[0]) and (
                0 <= global_col < room_mask.shape[1]
            ):
                # Check if the position is inside the room using the inner area mask
                if room_mask[global_row, global_col] == 1:
                    # If the position is the exception, consider it not inside the room
                    if exception and pos == exception:
                        continue
                    # Otherwise, it's inside this room
                    return True
        # If we reached here, the position isn't inside any room
        return False

    def generate_global_room_mask(self) -> np.ndarray:
        """
        Generate a global mask for the entire room.

        Returns:
            np.ndarray: The global room mask.
        """
        global_room_mask = np.zeros((self.rows, self.cols), dtype=bool)

        for room in self.rooms:
            local_mask = room.generate_inner_area_mask()
            global_position = room.global_position

            for i in range(local_mask.shape[0]):
                for j in range(local_mask.shape[1]):
                    global_room_mask[
                        global_position[0] + i, global_position[1] + j
                    ] |= local_mask[i, j]

        return global_room_mask

    #### Validate maze ####
    def is_valid_maze(self) -> bool:
        """
        Checks if the maze is valid.

        Returns:
            bool: True if the maze is valid, False otherwise.
        """

        start_point = self.start_position
        visited = set([start_point])
        queue = deque([start_point])

        while queue:
            current_point = queue.popleft()
            for neighbor in self.get_neighbors(*current_point):
                if neighbor not in visited:
                    visited.add(neighbor)
                    queue.append(neighbor)

        # Check if all PATH cells have been visited
        all_path_cells = {
            (x, y)
            for x in range(self.rows)
            for y in range(self.cols)
            if self.grid[x][y] == PATH
        }

        return visited == all_path_cells

    def get_neighbors(self, x: int, y: int) -> List[Tuple[int, int]]:
        """
        Given the coordinates (x, y), this function returns a list of neighboring coordinates.

        Parameters:
            x (int): The x-coordinate.
            y (int): The y-coordinate.

        Returns:
            List[Tuple[int, int]]: A list of neighboring coordinates, represented as tuples of (x, y).
        """
        neighbors = []
        for dx, dy in [(1, 0), (-1, 0), (0, 1), (0, -1)]:
            nx, ny = x + dx, y + dy
            if (
                0 <= nx < self.rows
                and 0 <= ny < self.cols
                and self.grid[nx][ny] == PATH
            ):
                neighbors.append((nx, ny))
        return neighbors



def on_line(p, q, r):
    """Check if point q lies on the line segment pr"""

    # 1. Check for collinearity
    area = (q[1] - p[1]) * (r[0] - q[0]) - (q[0] - p[0]) * (r[1] - q[1])
    if area != 0:
        return False

    # 2. Check if q is inside the bounding box
    if (
        q[0] <= max(p[0], r[0])
        and q[0] >= min(p[0], r[0])
        and q[1] <= max(p[1], r[1])
        and q[1] >= min(p[1], r[1])
    ):
        return True

    return False
